/* get.c - Get records.
   Copyright (C) 2004 Neal H. Walfield
   Written by Neal H. Walfield <neal@cs.uml.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <argp.h>

#include <db.h>

#include <assert.h>

#include "lib.h"
#include "database.h"

static char doc[] = "db get -- database shell utilities";
static char args_doc[] = "-d DATABASE -k KEY1 [-k KEY1...]";

static struct argp_option options [] = {
  { "database",		'd', "DATABASE",0, "Database to read" },
  { "key",		'k', "KEY",	0, "Key name" },
  { "key-from-file",	'K', "FILE",	0,
    "Use the contents of the file as the key's name" },
  { "assume-empty",	'e', 0,		0,
    "If a key is not found, pretend that it has an empty value" },
  { 0 }
};

static char *database;
static int assume_empty;

struct key
{
  char *key;
  int key_is_file;

  struct key *next;
};

static struct key *keys;
static struct key **last_nextp = &keys;

static error_t
parse_options (int key, char *arg, struct argp_state *state)
{
  switch (key)
    {
    case 'd':
      database = arg;
      break;

    case 'k':
    case 'K':
      {
	struct key *k = xmalloc (sizeof (struct key));

	k->key = arg;
	k->key_is_file = key == 'K';
	k->next = 0;

	*last_nextp = k;
	last_nextp = &k->next;

	break;
      }

    case 'e':
      assume_empty = 1;
      break;

    case ARGP_KEY_ARG:
      /* Too many arguments.  */
      argp_error (state, "too many arguments");

    case ARGP_KEY_END:
      if (! database)
	argp_error (state, "no database specified.");

      if (! keys)
	argp_error (state, "no keys to get.");

      return 0;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_options, args_doc, doc };

int
dbsh_get (int argc, char *argv[])
{
  int err;
  int index;
  DB *db;
  DBT key, value;
  char *separator = getenv ("IFS") ?: " ";
  int fail = 0;

  err = argp_parse (&argp, argc, argv, 0, &index, 0);
  if (err)
    error (1, err, "argp_parse");

  err = database_open (&db, database);
  if (err)
    return 1;

  memset(&key, 0, sizeof(key));
  memset(&value, 0, sizeof(value));
  value.flags = DB_DBT_REALLOC;

  for (struct key *k = keys; k; k = k->next)
    {
      int fd;

      if (k->key_is_file)
	{
	  fd = open (k->key, O_RDONLY);
	  if (fd == -1)
	    {
	      fail = 1;
	      error (0, errno, "%s", k->key);
	      continue;
	    }

	  key.size = lseek (fd, 0, SEEK_END);
	  key.data = mmap (0, key.size, PROT_READ, MAP_PRIVATE, fd, 0);
	  if (key.data == (void *) -1)
	    {
	      fail = 1;
	      error (0, errno, "%s", k->key);
	      continue;
	    }
	}
      else
	{
	  key.data = k->key;
	  key.size = strlen (k->key);
	}

      err = db->get (db, NULL, &key, &value, 0);
      switch (err)
	{
	case 0:
	  {
	    int i;
	    for (i = 0; i < value.size; i ++)
	      fputc (((char *) value.data)[i], stdout);
	    break;
	  }

	case DB_NOTFOUND:
	  if (assume_empty)
	    break;

	  /* Fall through.  */
	default:
	  db->err(db, err, "get (%*s)", key.size, key.data);
	  fail = 1;
	  break;
	}

      if (k->key_is_file)
	close (fd);

      /* Print the separator iff this is not the last element to
	 print.  */
      if (k->next)
	printf ("%s", separator);
    }

  database_close (db);

  return fail ? 1 : 0;
}
