/* put.c - Create new records.
   Copyright (C) 2004 Neal H. Walfield
   Written by Neal H. Walfield <neal@cs.uml.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <argp.h>

#include <db.h>

#include <assert.h>

#include "lib.h"
#include "database.h"

static char doc[] = "db put -- database shell utilities";
static char args_doc[] = "FILENAME -k KEY1 -v VALUE1 [-k KEY1 -vVALUE2...]";

static struct argp_option options [] = {
  { "database",		'd', "DATABASE",0, "Database to read" },
  { "key",	'k', "KEY",		0, "Key name" },
  { "key-from-file", 'K', "FILE",	0,
    "Use the contents of the file as the key's name" },
  { "value",	'v', "VALUE",		0, "Value" },
  { "value-from-file", 'V', "FILE",	0,
    "Use the contents of the file as the value" },
  { "replace",	'r', 0,		0,
    "if key is already present in the database overwrite it"},
  { 0 }
};

static char *database;
static int replace;

struct pair
{
  char *key;
  int key_is_file;
  char *value;
  int value_is_file;

  struct pair *next;
};

static struct pair *pairs;
static struct pair **last_nextp = &pairs;

static error_t
parse_options (int key, char *arg, struct argp_state *state)
{
  static struct pair *need_value;

  if (need_value && (key != 'v' && key != 'V'))
    argp_error (state, "key `%s' lacks corresponding value.",
		pairs->key);

  switch (key)
    {
    case 'd':
      database = arg;
      break;

    case 'k':
    case 'K':
      {
	struct pair *p = xmalloc (sizeof (struct pair));
	need_value = p;

	p->key = arg;
	p->key_is_file = key == 'K';
	p->next = 0;

	*last_nextp = p;
	last_nextp = &p->next;

	break;
      }

    case 'v':
    case 'V':
      assert (need_value);
      need_value->value = arg;
      need_value->value_is_file = key == 'V';
      need_value = 0;

      break;

    case 'r':
      replace = 1;
      break;

    case ARGP_KEY_ARG:
      /* Too many arguments.  */
      argp_error (state, "too many arguments");

    case ARGP_KEY_END:
      if (need_value)
	argp_error (state, "final key has no corresponding value to set.");

      if (! database)
	argp_error (state, "no database specified.");

      if (! pairs)
	argp_error (state, "no key/value pairs to put.");

      return 0;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_options, args_doc, doc };

int
dbsh_put (int argc, char *argv[])
{
  int err;
  int index;
  DB *db;
  DBT key, value;
  int fail = 0;
  int could_replace = 0;

  err = argp_parse (&argp, argc, argv, 0, &index, 0);
  if (err)
    error (1, err, "argp_parse");

  err = database_open (&db, database);
  if (err)
    return 1;

  for (struct pair *p = pairs; p; p = p->next)
    {
      int kfd;
      int vfd;

      memset(&key, 0, sizeof(key));
      if (p->key_is_file)
	{
	  kfd = open (p->key, O_RDONLY);
	  if (kfd == -1)
	    {
	      error (0, errno, "%s", p->key);
	      fail = 1;
	      continue;
	    }

	  key.size = lseek (kfd, 0, SEEK_END);
	  key.data = mmap (0, key.size, PROT_READ, MAP_PRIVATE, kfd, 0);
	  if (key.data == (void *) -1)
	    {
	      error (0, errno, "%s", p->key);
	      fail = 1;
	      continue;
	    }
	}
      else
	{
	  key.data = p->key;
	  key.size = strlen (p->key);
	}

      memset(&value, 0, sizeof(value));
      if (p->value_is_file)
	{
	  vfd = open (p->value, O_RDONLY);
	  if (vfd == -1)
	    {
	      error (0, errno, "%s", p->value);
	      fail = 1;
	      continue;
	    }

	  value.size = lseek (vfd, 0, SEEK_END);
	  value.data = mmap (0, value.size, PROT_READ, MAP_PRIVATE, vfd, 0);
	  if (value.data == (void *) -1)
	    {
	      error (0, errno, "%s", p->value);
	      fail = 1;
	      continue;
	    }
	}
      else
	{
	  value.data = p->value;
	  value.size = strlen (p->value);
	}

      err = db->put (db, NULL, &key, &value, replace ? 0 : DB_NOOVERWRITE);
      if (err)
	{
	  if (err == DB_KEYEXIST)
	    could_replace = 1;
	  else
	    fail = 1;
 
	  db->err(db, err, "put (%s, %s)", p->key, p->value);
	}

      if (p->key_is_file)
	close (kfd);
      if (p->value_is_file)
	close (vfd);
    }

  database_close (db);

  return (could_replace ? 2 : 0) | (fail ? 1 : 0);
}
