/* search.c - Search for records.
   Copyright (C) 2004 Neal H. Walfield
   Written by Neal H. Walfield <neal@cs.uml.edu>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <regex.h>
#include <errno.h>
#include <error.h>
#include <argp.h>

#include <db.h>

#include <assert.h>

#include "lib.h"
#include "database.h"

#include "search.h"
#include "search-exp.h"

static char doc[] = "db search -- database shell utilities\n\
If no patterns are specified, prints all records.";
static char args_doc[] = "-d DATABASE [PATTERN...]";

static struct argp_option options [] = {
  { "database",		'd', "DATABASE",0, "Database to read" },
  { "key",		'k', "KEY-RE",	0, "Only apply RE to keys" },
  { "value",		'v', "VALUE-RE",0, "Only apply RE to values" },
  { "and",		'a', 0, 0, "Require the conjunction of the RE" },
  { "or",		'o', 0, 0, "Require the disjunction of the RE" },
  { "not",		'!', 0, 0, "Invert the next result" },
  { "(",		'(', 0, 0, "Start group" },
  { ")",		')', 0, 0, "End group" },
  { "suppress-keys",	1, 0, 0,	   "Don't print keys" },
  { "suppress-values",  2, 0, 0,	   "Don't print values" },
  { 0 }
};

static char *database;
static int suppress_keys;
static int suppress_values;
static char *separator;

struct op
{
  /* 'a' -> and; 'o' -> or; '!' -> not; '(' -> left paren; ')' -> right
     'p' -> pattern.  */
  int type;
  struct pattern *pattern;
  struct op *next;
};

static struct op *ops;
static struct op **last_nextp = &ops;

int
yylex ()
{
  int ret;
  struct op *t;

  if (! ops)
    /* EOF.  */
    return 0;

  ret = ops->type;
  if (ret == 'p')
    {
      ret = PATTERN;
      yylval.pattern = ops->pattern;
    }

  t = ops->next;
  free (ops);
  ops = t;

  return ret;
}

static error_t
parse_options (int key, char *arg, struct argp_state *state)
{
  switch (key)
    {
    case 1:
      suppress_keys = 1;
      break;
    case 2:
      suppress_values = 1;
      break;

    case 'd':
      database = arg;
      break;

    case 'a':
    case 'o':
    case '!':
    case '(':
    case ')':
      {
	struct op *op = xcalloc (sizeof (struct op), 1);

	op->type = key;

	*last_nextp = op;
	last_nextp = &op->next;

	break;
      }

    case 'k':
    case 'v':
    case ARGP_KEY_ARG:
      {
	int err;
	struct op *op = xcalloc (sizeof (struct op), 1);
	struct pattern *p = xmalloc (sizeof (struct pattern));

	op->type = 'p';
	op->pattern = p;

	p->pattern = arg;
	p->apply_to_key = key == 'k' || key == ARGP_KEY_ARG;
	p->apply_to_value = key == 'v' || key == ARGP_KEY_ARG;

	err = regcomp (&p->re, p->pattern, REG_NOSUB);
	if (err)
	  {
	    int len = regerror (err, &p->re, 0, 0);
	    char *b;
	    if (len < 1024)
	      b = alloca (len);
	    else
	      b = xmalloc (len);

	    regerror (err, &p->re, b, len);
	    
	    error (1, 0, "compiling regular expression `%s': %s",
		   p->pattern, b);
	  }

	*last_nextp = op;
	last_nextp = &op->next;

	break;
      }

    case ARGP_KEY_END:
      if (! database)
	argp_error (state, "no database specified.");

      if (ops)
	{
	  extern int yyparse (void);
	  int err;

#ifdef DEBUG
	  {
	    struct op *op;
	    for (op = ops; op; op = op->next)
	      {
		printf ("%c ", op->type);
		if (op->type == 'p')
		  printf ("(%s) ", op->pattern->pattern);
		printf ("\n");
	      }
	  }
#endif /* DEBUG */

	  err = yyparse ();
	  if (err)
	    error (1, 0, "Invalid argument syntax");

#ifdef DEBUG
	  /* Dump the parse tree.  */
	  {
	    void dump (struct expression *e)
	      {
		switch (e->type)
		  {
		  case exp_pattern:
		    printf ("%s/%s/", e->invert_logic ? "!" : "",
			    e->pattern.pattern->pattern);
		    break;

		  case exp_and:
		    printf ("%sand (", e->invert_logic ? "!" : "");
		    /* fall through.  */
		  case exp_or:
		    if (e->type == exp_or)
		      printf ("%sor (", e->invert_logic ? "!" : "");

		    dump (e->expression.a);
		    printf (", ");
		    dump (e->expression.b);
		    printf (")");
		    break;

		  default:
		    assert (! "Internal inconsistency");
		  }
	      }

	    dump (expressions);
	    printf ("\n");
	  }
#endif /* DEBUG  */
	}

      return 0;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_options, args_doc, doc };

/* Do we print a key?  Test it against the expression tree and if so
   print it.  */
static void
eval (char *key_data, int key_len, char *value_data, int value_len)
{
  char *k = 0, *v = 0;

  int test_pattern (struct pattern *p)
    {
      int err;

      if (p->apply_to_key)
	{
	  if (! k)
	    {
	      k = xmalloc (key_len + 1);
	      memcpy (k, key_data, key_len);
	      k[key_len] = 0;
	    }

	  err = regexec (&p->re, k, 0, 0, 0);
	  if (! err)
	    return 1;
	}

      if (p->apply_to_value)
	{
	  if (! v)
	    {
	      v = xmalloc (value_len + 1);
	      memcpy (v, value_data, value_len);
	      v[value_len] = 0;
	    }

	  err = regexec (&p->re, v, 0, 0, 0);
	  if (! err)
	    return 1;
	}

      return 0;
    }

  int test (struct expression *exp)
    {
      int res;

      switch (exp->type)
	{
	case exp_pattern:
	  res = test_pattern (exp->pattern.pattern);
	  break;

	case exp_or:
	  res = test (exp->expression.a)
	    || test (exp->expression.b);
	  break;

	case exp_and:
	  res = test (exp->expression.a)
	    && test (exp->expression.b);
	  break;

	default:
	  assert (! "Fatal error: Internal consistency error.");
	}

      if (exp->invert_logic)
	return ! res;
      return res;
    }

  int matched = 1;
  if (expressions)
    matched = test (expressions);

  if (matched && ! suppress_keys)
    {
      if (k)
	fputs (k, stdout);
      else
	{
	  int i;
	  for (i = 0; i < key_len; i ++)
	    fputc (key_data[i], stdout);
	}

      fputs (separator, stdout);
    }

  if (matched && ! suppress_values)
    {
      if (v)
	fputs (v, stdout);
      else
	{
	  int i;
	  for (i = 0; i < value_len; i ++)
	    fputc (value_data[i], stdout);
	  fputs (separator, stdout);
	}
    }

  if (k)
    free (k);
  if (v)
    free (v);
}

int
dbsh_search (int argc, char *argv[])
{
  int err;
  DB *db;
  DBC *cursor;
  DBT key, value;
  size_t key_len, value_len;
  char *key_data, *value_data;
  void *p;

  int fail = 0;

  separator = getenv ("IFS") ?: "\n";

  err = argp_parse (&argp, argc, argv, ARGP_IN_ORDER, 0, 0);
  if (err)
    error (1, err, "argp_parse");

  err = database_open (&db, database);
  if (err)
    return 1;

  memset (&key, 0, sizeof (key));
  memset (&value, 0, sizeof (value));

  /* Review the database in 5MB chunks. */
#define BUFFER_LENGTH (5 * 1024 * 1024)
  value.data = xmalloc (BUFFER_LENGTH);
  value.ulen = BUFFER_LENGTH; 
  value.flags = DB_DBT_USERMEM;

  /* Acquire a cursor.  */
  err = db->cursor (db, NULL, &cursor, 0);
  if (err)
    {
      db->err (db, err, "DB->cursor"); 
      free (value.data); 
      return 1;
    }

  while (1)
    {
      /* Acquire the next set of key/data pairs. This code does not
	 handle single key/data pairs that won't fit in a
	 BUFFER_LENGTH size buffer, instead returning ENOMEM to our
	 caller. */
      err = cursor->c_get (cursor, &key, &value,
			   DB_MULTIPLE_KEY | DB_NEXT);
      if (err)
	{
	  if (err != DB_NOTFOUND)
	    {
	      fail = 1;
	      db->err (db, err, "DBcursor->c_get");
	    }
	  break;
	}

      DB_MULTIPLE_INIT (p, &value);
      while (1)
	{ 
	  DB_MULTIPLE_KEY_NEXT (p, &value,
				key_data, key_len, value_data, value_len);
	  if (! p)
	    break;

	  eval (key_data, key_len, value_data, value_len);
	}
    }

  err = cursor->c_close(cursor);
  if (err)
    {
      fail = 1;
      db->err(db, err, "DBcursor->close");
    }

  database_close (db);

  return fail ? 1 : 0;
}
